@@ -1,3 +1,6 @@
+// SPDX-FileCopyrightText: 2023 The Pion community <https://pion.ly>
+// SPDX-License-Identifier: MIT
+
 //go:build !js
 // +build !js
 
@@ -7,22 +10,20 @@ import (
 	"errors"
 	"fmt"
 	"io"
-	"math"
 	"sync"
 	"sync/atomic"
 	"time"
 
 	"github.com/pion/datachannel"
 	"github.com/pion/logging"
-	"github.com/pion/webrtc/v3/pkg/rtcerr"
+	"github.com/pion/webrtc/v4/pkg/rtcerr"
 )
 
-const dataChannelBufferSize = math.MaxUint16 // message size limit for Chromium
 var errSCTPNotEstablished = errors.New("SCTP not established")
 
 // DataChannel represents a WebRTC DataChannel
 // The DataChannel interface represents a network channel
-// which can be used for bidirectional peer-to-peer transfers of arbitrary data
+// which can be used for bidirectional peer-to-peer transfers of arbitrary data.
 type DataChannel struct {
 	mu sync.RWMutex
 
@@ -37,6 +38,8 @@ type DataChannel struct {
 	readyState                 atomic.Value // DataChannelState
 	bufferedAmountLowThreshold uint64
 	detachCalled               bool
+	readLoopActive             chan struct{}
+	isGracefulClosed           bool
 
 	// The binaryType represents attribute MUST, on getting, return the value to
 	// which it was last set. On setting, if the new value is either the string
@@ -67,7 +70,7 @@ type DataChannel struct {
 // This constructor is part of the ORTC API. It is not
 // meant to be used together with the basic WebRTC API.
 func (api *API) NewDataChannel(transport *SCTPTransport, params *DataChannelParameters) (*DataChannel, error) {
-	d, err := api.newDataChannel(params, api.settingEngine.LoggerFactory.NewLogger("ortc"))
+	d, err := api.newDataChannel(params, nil, api.settingEngine.LoggerFactory.NewLogger("ortc"))
 	if err != nil {
 		return nil, err
 	}
@@ -82,13 +85,18 @@ func (api *API) NewDataChannel(transport *SCTPTransport, params *DataChannelPara
 
 // newDataChannel is an internal constructor for the data channel used to
 // create the DataChannel object before the networking is set up.
-func (api *API) newDataChannel(params *DataChannelParameters, log logging.LeveledLogger) (*DataChannel, error) {
+func (api *API) newDataChannel(
+	params *DataChannelParameters,
+	sctpTransport *SCTPTransport,
+	log logging.LeveledLogger,
+) (*DataChannel, error) {
 	// https://w3c.github.io/webrtc-pc/#peer-to-peer-data-api (Step #5)
 	if len(params.Label) > 65535 {
 		return nil, &rtcerr.TypeError{Err: ErrStringSizeLimit}
 	}
 
-	d := &DataChannel{
+	dataChannel := &DataChannel{
+		sctpTransport:     sctpTransport,
 		statsID:           fmt.Sprintf("DataChannel-%d", time.Now().UnixNano()),
 		label:             params.Label,
 		protocol:          params.Protocol,
@@ -101,12 +109,13 @@ func (api *API) newDataChannel(params *DataChannelParameters, log logging.Levele
 		log:               log,
 	}
 
-	d.setReadyState(DataChannelStateConnecting)
-	return d, nil
+	dataChannel.setReadyState(DataChannelStateConnecting)
+
+	return dataChannel, nil
 }
 
-// open opens the datachannel over the sctp transport
-func (d *DataChannel) open(sctpTransport *SCTPTransport) error {
+// open opens the datachannel over the sctp transport.
+func (d *DataChannel) open(sctpTransport *SCTPTransport) error { //nolint:cyclop
 	association := sctpTransport.association()
 	if association == nil {
 		return errSCTPNotEstablished
@@ -115,6 +124,7 @@ func (d *DataChannel) open(sctpTransport *SCTPTransport) error {
 	d.mu.Lock()
 	if d.sctpTransport != nil { // already open
 		d.mu.Unlock()
+
 		return nil
 	}
 	d.sctpTransport = sctpTransport
@@ -169,6 +179,7 @@ func (d *DataChannel) open(sctpTransport *SCTPTransport) error {
 	dc, err := datachannel.Dial(association, *d.id, cfg)
 	if err != nil {
 		d.mu.Unlock()
+
 		return err
 	}
 
@@ -179,6 +190,7 @@ func (d *DataChannel) open(sctpTransport *SCTPTransport) error {
 
 	d.onDial()
 	d.handleOpen(dc, false, d.negotiated)
+
 	return nil
 }
 
@@ -191,7 +203,7 @@ func (d *DataChannel) Transport() *SCTPTransport {
 }
 
 // After onOpen is complete check that the user called detach
-// and provide an error message if the call was missed
+// and provide an error message if the call was missed.
 func (d *DataChannel) checkDetachAfterOpen() {
 	d.mu.RLock()
 	defer d.mu.RUnlock()
@@ -221,6 +233,11 @@ func (d *DataChannel) OnOpen(f func()) {
 func (d *DataChannel) onOpen() {
 	d.mu.RLock()
 	handler := d.onOpenHandler
+	if d.isGracefulClosed {
+		d.mu.RUnlock()
+
+		return
+	}
 	d.mu.RUnlock()
 
 	if handler != nil {
@@ -232,7 +249,7 @@ func (d *DataChannel) onOpen() {
 }
 
 // OnDial sets an event handler which is invoked when the
-// peer has been dialed, but before said peer has responsed
+// peer has been dialed, but before said peer has responded.
 func (d *DataChannel) OnDial(f func()) {
 	d.mu.Lock()
 	d.dialHandlerOnce = sync.Once{}
@@ -248,6 +265,11 @@ func (d *DataChannel) OnDial(f func()) {
 func (d *DataChannel) onDial() {
 	d.mu.RLock()
 	handler := d.onDialHandler
+	if d.isGracefulClosed {
+		d.mu.RUnlock()
+
+		return
+	}
 	d.mu.RUnlock()
 
 	if handler != nil {
@@ -257,6 +279,10 @@ func (d *DataChannel) onDial() {
 
 // OnClose sets an event handler which is invoked when
 // the underlying data transport has been closed.
+// Note: Due to backwards compatibility, there is a chance that
+// OnClose can be called, even if the GracefulClose is used.
+// If this is the case for you, you can deregister OnClose
+// prior to GracefulClose.
 func (d *DataChannel) OnClose(f func()) {
 	d.mu.Lock()
 	defer d.mu.Unlock()
@@ -288,6 +314,11 @@ func (d *DataChannel) OnMessage(f func(msg DataChannelMessage)) {
 func (d *DataChannel) onMessage(msg DataChannelMessage) {
 	d.mu.RLock()
 	handler := d.onMessageHandler
+	if d.isGracefulClosed {
+		d.mu.RUnlock()
+
+		return
+	}
 	d.mu.RUnlock()
 
 	if handler == nil {
@@ -298,7 +329,18 @@ func (d *DataChannel) onMessage(msg DataChannelMessage) {
 
 func (d *DataChannel) handleOpen(dc *datachannel.DataChannel, isRemote, isAlreadyNegotiated bool) {
 	d.mu.Lock()
+	if d.isGracefulClosed { // The channel was closed during the connecting state
+		d.mu.Unlock()
+		if err := dc.Close(); err != nil {
+			d.log.Errorf("Failed to close DataChannel that was closed during connecting state %v", err.Error())
+		}
+		d.onClose()
+
+		return
+	}
 	d.dataChannel = dc
+	bufferedAmountLowThreshold := d.bufferedAmountLowThreshold
+	onBufferedAmountLow := d.onBufferedAmountLow
 	d.mu.Unlock()
 	d.setReadyState(DataChannelStateOpen)
 
@@ -307,6 +349,9 @@ func (d *DataChannel) handleOpen(dc *datachannel.DataChannel, isRemote, isAlread
 	// * remote datachannels should fire OnOpened. This isn't spec compliant, but we can't break behavior yet
 	// * already negotiated datachannels should fire OnOpened
 	if d.api.settingEngine.detach.DataChannels || isRemote || isAlreadyNegotiated {
+		// bufferedAmountLowThreshold and onBufferedAmountLow might be set earlier
+		d.dataChannel.SetBufferedAmountLowThreshold(bufferedAmountLowThreshold)
+		d.dataChannel.OnBufferedAmountLow(onBufferedAmountLow)
 		d.onOpen()
 	} else {
 		dc.OnOpen(func() {
@@ -317,7 +362,12 @@ func (d *DataChannel) handleOpen(dc *datachannel.DataChannel, isRemote, isAlread
 	d.mu.Lock()
 	defer d.mu.Unlock()
 
+	if d.isGracefulClosed {
+		return
+	}
+
 	if !d.api.settingEngine.detach.DataChannels {
+		d.readLoopActive = make(chan struct{})
 		go d.readLoop()
 	}
 }
@@ -333,6 +383,11 @@ func (d *DataChannel) OnError(f func(err error)) {
 func (d *DataChannel) onError(err error) {
 	d.mu.RLock()
 	handler := d.onErrorHandler
+	if d.isGracefulClosed {
+		d.mu.RUnlock()
+
+		return
+	}
 	d.mu.RUnlock()
 
 	if handler != nil {
@@ -340,37 +395,48 @@ func (d *DataChannel) onError(err error) {
 	}
 }
 
-// See https://github.com/pion/webrtc/issues/1516
-// nolint:gochecknoglobals
-var rlBufPool = sync.Pool{New: func() interface{} {
-	return make([]byte, dataChannelBufferSize)
-}}
-
 func (d *DataChannel) readLoop() {
+	defer func() {
+		d.mu.Lock()
+		readLoopActive := d.readLoopActive
+		d.mu.Unlock()
+		defer close(readLoopActive)
+	}()
+
+	buffer := make([]byte, sctpMaxMessageSizeUnsetValue)
 	for {
-		buffer := rlBufPool.Get().([]byte) //nolint:forcetypeassert
 		n, isString, err := d.dataChannel.ReadDataChannel(buffer)
 		if err != nil {
-			rlBufPool.Put(buffer) // nolint:staticcheck
+			if errors.Is(err, io.ErrShortBuffer) {
+				if int64(n) < int64(d.api.settingEngine.getSCTPMaxMessageSize()) {
+					buffer = append(buffer, make([]byte, len(buffer))...) // nolint
+
+					continue
+				}
+
+				d.log.Errorf(
+					"Incoming DataChannel message larger then Max Message size %v",
+					d.api.settingEngine.getSCTPMaxMessageSize(),
+				)
+			}
+
 			d.setReadyState(DataChannelStateClosed)
 			if !errors.Is(err, io.EOF) {
 				d.onError(err)
 			}
 			d.onClose()
+
 			return
 		}
 
-		m := DataChannelMessage{Data: make([]byte, n), IsString: isString}
-		copy(m.Data, buffer[:n])
-		// The 'staticcheck' pragma is a false positive on the part of the CI linter.
-		rlBufPool.Put(buffer) // nolint:staticcheck
-
-		// NB: Why was DataChannelMessage not passed as a pointer value?
-		d.onMessage(m) // nolint:staticcheck
+		d.onMessage(DataChannelMessage{
+			Data:     append([]byte{}, buffer[:n]...),
+			IsString: isString,
+		})
 	}
 }
 
-// Send sends the binary message to the DataChannel peer
+// Send sends the binary message to the DataChannel peer.
 func (d *DataChannel) Send(data []byte) error {
 	err := d.ensureOpen()
 	if err != nil {
@@ -378,10 +444,11 @@ func (d *DataChannel) Send(data []byte) error {
 	}
 
 	_, err = d.dataChannel.WriteDataChannel(data, false)
+
 	return err
 }
 
-// SendText sends the text message to the DataChannel peer
+// SendText sends the text message to the DataChannel peer.
 func (d *DataChannel) SendText(s string) error {
 	err := d.ensureOpen()
 	if err != nil {
@@ -389,6 +456,7 @@ func (d *DataChannel) SendText(s string) error {
 	}
 
 	_, err = d.dataChannel.WriteDataChannel([]byte(s), true)
+
 	return err
 }
 
@@ -398,11 +466,15 @@ func (d *DataChannel) ensureOpen() error {
 	if d.ReadyState() != DataChannelStateOpen {
 		return io.ErrClosedPipe
 	}
+
 	return nil
 }
 
-// Detach allows you to detach the underlying datachannel. This provides
-// an idiomatic API to work with, however it disables the OnMessage callback.
+// Detach allows you to detach the underlying datachannel.
+// This provides an idiomatic API to work with
+// (`io.ReadWriteCloser` with its `.Read()` and `.Write()` methods,
+// as opposed to `.Send()` and `.OnMessage`),
+// however it disables the OnMessage callback.
 // Before calling Detach you have to enable this behavior by calling
 // webrtc.DetachDataChannels(). Combining detached and normal data channels
 // is not supported.
@@ -410,26 +482,81 @@ func (d *DataChannel) ensureOpen() error {
 // pion/datachannel documentation for the correct way to handle the
 // resulting DataChannel object.
 func (d *DataChannel) Detach() (datachannel.ReadWriteCloser, error) {
+	return d.DetachWithDeadline()
+}
+
+// DetachWithDeadline allows you to detach the underlying datachannel.
+// It is the same as Detach but returns a ReadWriteCloserDeadliner.
+func (d *DataChannel) DetachWithDeadline() (datachannel.ReadWriteCloserDeadliner, error) {
 	d.mu.Lock()
-	defer d.mu.Unlock()
 
 	if !d.api.settingEngine.detach.DataChannels {
+		d.mu.Unlock()
+
 		return nil, errDetachNotEnabled
 	}
 
 	if d.dataChannel == nil {
+		d.mu.Unlock()
+
 		return nil, errDetachBeforeOpened
 	}
 
 	d.detachCalled = true
 
-	return d.dataChannel, nil
+	dataChannel := d.dataChannel
+	d.mu.Unlock()
+
+	// Remove the reference from SCTPTransport so that the datachannel
+	// can be garbage collected on close
+	d.sctpTransport.lock.Lock()
+	n := len(d.sctpTransport.dataChannels)
+	j := 0
+	for i := 0; i < n; i++ {
+		if d == d.sctpTransport.dataChannels[i] {
+			continue
+		}
+		d.sctpTransport.dataChannels[j] = d.sctpTransport.dataChannels[i]
+		j++
+	}
+	for i := j; i < n; i++ {
+		d.sctpTransport.dataChannels[i] = nil
+	}
+	d.sctpTransport.dataChannels = d.sctpTransport.dataChannels[:j]
+	d.sctpTransport.lock.Unlock()
+
+	return dataChannel, nil
 }
 
 // Close Closes the DataChannel. It may be called regardless of whether
 // the DataChannel object was created by this peer or the remote peer.
 func (d *DataChannel) Close() error {
+	return d.close(false)
+}
+
+// GracefulClose Closes the DataChannel. It may be called regardless of whether
+// the DataChannel object was created by this peer or the remote peer. It also waits
+// for any goroutines it started to complete. This is only safe to call outside of
+// DataChannel callbacks or if in a callback, in its own goroutine.
+func (d *DataChannel) GracefulClose() error {
+	return d.close(true)
+}
+
+// Normally, close only stops writes from happening, so graceful=true
+// will wait for reads to be finished based on underlying SCTP association
+// closure or a SCTP reset stream from the other side. This is safe to call
+// with graceful=true after tearing down a PeerConnection but not
+// necessarily before. For example, if you used a vnet and dropped all packets
+// right before closing the DataChannel, you'd need never see a reset stream.
+func (d *DataChannel) close(shouldGracefullyClose bool) error {
 	d.mu.Lock()
+	d.isGracefulClosed = true
+	readLoopActive := d.readLoopActive
+	if shouldGracefullyClose && readLoopActive != nil {
+		defer func() {
+			<-readLoopActive
+		}()
+	}
 	haveSctpTransport := d.dataChannel != nil
 	d.mu.Unlock()
 
@@ -518,6 +645,7 @@ func (d *DataChannel) ReadyState() DataChannelState {
 	if v, ok := d.readyState.Load().(DataChannelState); ok {
 		return v
 	}
+
 	return DataChannelState(0)
 }
 
@@ -538,6 +666,7 @@ func (d *DataChannel) BufferedAmount() uint64 {
 	if d.dataChannel == nil {
 		return 0
 	}
+
 	return d.dataChannel.BufferedAmount()
 }
 
@@ -554,6 +683,7 @@ func (d *DataChannel) BufferedAmountLowThreshold() uint64 {
 	if d.dataChannel == nil {
 		return d.bufferedAmountLowThreshold
 	}
+
 	return d.dataChannel.BufferedAmountLowThreshold()
 }
 
@@ -571,21 +701,36 @@ func (d *DataChannel) SetBufferedAmountLowThreshold(th uint64) {
 }
 
 // OnBufferedAmountLow sets an event handler which is invoked when
-// the number of bytes of outgoing data becomes lower than the
+// the number of bytes of outgoing data becomes lower than or equal to the
 // BufferedAmountLowThreshold.
 func (d *DataChannel) OnBufferedAmountLow(f func()) {
 	d.mu.Lock()
 	defer d.mu.Unlock()
 
-	d.onBufferedAmountLow = f
+	onBufferedAmountLow := d.makeBufferedAmountLowHandler(f)
+	d.onBufferedAmountLow = onBufferedAmountLow
+
 	if d.dataChannel != nil {
-		d.dataChannel.OnBufferedAmountLow(f)
+		d.dataChannel.OnBufferedAmountLow(onBufferedAmountLow)
+	}
+}
+
+func (d *DataChannel) makeBufferedAmountLowHandler(f func()) func() {
+	return func() {
+		go func() {
+			if d.ReadyState() != DataChannelStateOpen {
+				return
+			}
+
+			f()
+		}()
 	}
 }
 
 func (d *DataChannel) getStatsID() string {
 	d.mu.Lock()
 	defer d.mu.Unlock()
+
 	return d.statsID
 }
 
