@@ -1,13 +1,17 @@
+// SPDX-FileCopyrightText: 2023 The Pion community <https://pion.ly>
+// SPDX-License-Identifier: MIT
+
 package webrtc
 
 import (
 	"fmt"
 	"io"
 	"sync"
+	"sync/atomic"
 	"testing"
 	"time"
 
-	"github.com/pion/transport/v2/test"
+	"github.com/pion/transport/v3/test"
 	"github.com/stretchr/testify/assert"
 )
 
@@ -16,50 +20,55 @@ import (
 // bindings this is a requirement).
 const expectedLabel = "data"
 
-func closePairNow(t testing.TB, pc1, pc2 io.Closer) {
+func closePairNow(tb testing.TB, pc1, pc2 io.Closer) {
+	tb.Helper()
+
 	var fail bool
 	if err := pc1.Close(); err != nil {
-		t.Errorf("Failed to close PeerConnection: %v", err)
+		tb.Errorf("Failed to close PeerConnection: %v", err)
 		fail = true
 	}
 	if err := pc2.Close(); err != nil {
-		t.Errorf("Failed to close PeerConnection: %v", err)
+		tb.Errorf("Failed to close PeerConnection: %v", err)
 		fail = true
 	}
 	if fail {
-		t.FailNow()
+		tb.FailNow()
 	}
 }
 
 func closePair(t *testing.T, pc1, pc2 io.Closer, done <-chan bool) {
+	t.Helper()
+
 	select {
 	case <-time.After(10 * time.Second):
-		t.Fatalf("closePair timed out waiting for done signal")
+		assert.Fail(t, "closePair timed out waiting for done signal")
 	case <-done:
 		closePairNow(t, pc1, pc2)
 	}
 }
 
-func setUpDataChannelParametersTest(t *testing.T, options *DataChannelInit) (*PeerConnection, *PeerConnection, *DataChannel, chan bool) {
+func setUpDataChannelParametersTest(
+	t *testing.T,
+	options *DataChannelInit,
+) (*PeerConnection, *PeerConnection, *DataChannel, chan bool) {
+	t.Helper()
+
 	offerPC, answerPC, err := newPair()
-	if err != nil {
-		t.Fatalf("Failed to create a PC pair for testing")
-	}
+	assert.NoError(t, err)
 	done := make(chan bool)
 
 	dc, err := offerPC.CreateDataChannel(expectedLabel, options)
-	if err != nil {
-		t.Fatalf("Failed to create a PC pair for testing")
-	}
+	assert.NoError(t, err)
 
 	return offerPC, answerPC, dc, done
 }
 
 func closeReliabilityParamTest(t *testing.T, pc1, pc2 *PeerConnection, done chan bool) {
+	t.Helper()
+
 	err := signalPair(pc1, pc2)
-	if err != nil {
-		t.Fatalf("Failed to signal our PC pair for testing")
-	}
+	assert.NoError(t, err)
 
 	closePair(t, pc1, pc2, done)
 }
@@ -72,6 +81,8 @@ func BenchmarkDataChannelSend32(b *testing.B) { benchmarkDataChannelSend(b, 32)
 
 // See https://github.com/pion/webrtc/issues/1516
 func benchmarkDataChannelSend(b *testing.B, numChannels int) {
+	b.Helper()
+
 	offerPC, answerPC, err := newPair()
 	if err != nil {
 		b.Fatalf("Failed to create a PC pair for testing")
@@ -118,9 +129,7 @@ func TestDataChannel_Open(t *testing.T) {
 		defer report()
 
 		offerPC, answerPC, err := newPair()
-		if err != nil {
-			t.Fatalf("Failed to create a PC pair for testing")
-		}
+		assert.NoError(t, err)
 
 		done := make(chan bool)
 		openCalls := make(chan bool, openOnceChannelCapacity)
@@ -132,7 +141,7 @@ func TestDataChannel_Open(t *testing.T) {
 			d.OnOpen(func() {
 				openCalls <- true
 			})
-			d.OnMessage(func(msg DataChannelMessage) {
+			d.OnMessage(func(DataChannelMessage) {
 				go func() {
 					// Wait a little bit to ensure all messages are processed.
 					time.Sleep(100 * time.Millisecond)
@@ -145,10 +154,7 @@ func TestDataChannel_Open(t *testing.T) {
 		assert.NoError(t, err)
 
 		dc.OnOpen(func() {
-			e := dc.SendText("Ping")
-			if e != nil {
-				t.Fatalf("Failed to send string on data channel")
-			}
+			assert.NoError(t, dc.SendText("Ping"), "Failed to send string on data channel")
 		})
 
 		assert.NoError(t, signalPair(offerPC, answerPC))
@@ -163,9 +169,7 @@ func TestDataChannel_Open(t *testing.T) {
 		defer report()
 
 		offerPC, answerPC, err := newPair()
-		if err != nil {
-			t.Fatalf("Failed to create a PC pair for testing")
-		}
+		assert.NoError(t, err)
 
 		done := make(chan bool)
 		answerOpenCalls := make(chan bool, openOnceChannelCapacity)
@@ -188,7 +192,7 @@ func TestDataChannel_Open(t *testing.T) {
 		})
 		assert.NoError(t, err)
 
-		answerDC.OnMessage(func(msg DataChannelMessage) {
+		answerDC.OnMessage(func(DataChannelMessage) {
 			go func() {
 				// Wait a little bit to ensure all messages are processed.
 				time.Sleep(100 * time.Millisecond)
@@ -201,10 +205,7 @@ func TestDataChannel_Open(t *testing.T) {
 
 		offerDC.OnOpen(func() {
 			offerOpenCalls <- true
-			e := offerDC.SendText("Ping")
-			if e != nil {
-				t.Fatalf("Failed to send string on data channel")
-			}
+			assert.NoError(t, offerDC.SendText("Ping"), "Failed to send string on data channel")
 		})
 
 		assert.NoError(t, signalPair(offerPC, answerPC))
@@ -216,15 +217,13 @@ func TestDataChannel_Open(t *testing.T) {
 	})
 }
 
-func TestDataChannel_Send(t *testing.T) {
+func TestDataChannel_Send(t *testing.T) { //nolint:cyclop
 	t.Run("before signaling", func(t *testing.T) {
 		report := test.CheckRoutines(t)
 		defer report()
 
 		offerPC, answerPC, err := newPair()
-		if err != nil {
-			t.Fatalf("Failed to create a PC pair for testing")
-		}
+		assert.NoError(t, err)
 
 		done := make(chan bool)
 
@@ -234,36 +233,26 @@ func TestDataChannel_Send(t *testing.T) {
 			if d.Label() != expectedLabel {
 				return
 			}
-			d.OnMessage(func(msg DataChannelMessage) {
-				e := d.Send([]byte("Pong"))
-				if e != nil {
-					t.Fatalf("Failed to send string on data channel")
-				}
+			d.OnMessage(func(DataChannelMessage) {
+				assert.NoError(t, d.Send([]byte("Pong")), "Failed to send string on data channel")
 			})
 			assert.True(t, d.Ordered(), "Ordered should be set to true")
 		})
 
 		dc, err := offerPC.CreateDataChannel(expectedLabel, nil)
-		if err != nil {
-			t.Fatalf("Failed to create a PC pair for testing")
-		}
+		assert.NoError(t, err)
 
 		assert.True(t, dc.Ordered(), "Ordered should be set to true")
 
 		dc.OnOpen(func() {
-			e := dc.SendText("Ping")
-			if e != nil {
-				t.Fatalf("Failed to send string on data channel")
-			}
+			assert.NoError(t, dc.SendText("Ping"), "Failed to send string on data channel")
 		})
-		dc.OnMessage(func(msg DataChannelMessage) {
+		dc.OnMessage(func(DataChannelMessage) {
 			done <- true
 		})
 
 		err = signalPair(offerPC, answerPC)
-		if err != nil {
-			t.Fatalf("Failed to signal our PC pair for testing: %+v", err)
-		}
+		assert.NoError(t, err)
 
 		closePair(t, offerPC, answerPC, done)
 	})
@@ -273,9 +262,7 @@ func TestDataChannel_Send(t *testing.T) {
 		defer report()
 
 		offerPC, answerPC, err := newPair()
-		if err != nil {
-			t.Fatalf("Failed to create a PC pair for testing")
-		}
+		assert.NoError(t, err)
 
 		done := make(chan bool)
 
@@ -285,11 +272,8 @@ func TestDataChannel_Send(t *testing.T) {
 			if d.Label() != expectedLabel {
 				return
 			}
-			d.OnMessage(func(msg DataChannelMessage) {
-				e := d.Send([]byte("Pong"))
-				if e != nil {
-					t.Fatalf("Failed to send string on data channel")
-				}
+			d.OnMessage(func(DataChannelMessage) {
+				assert.NoError(t, d.Send([]byte("Pong")), "Failed to send string on data channel")
 			})
 			assert.True(t, d.Ordered(), "Ordered should be set to true")
 		})
@@ -300,23 +284,18 @@ func TestDataChannel_Send(t *testing.T) {
 				// wasm fires completed state multiple times
 				once.Do(func() {
 					dc, createErr := offerPC.CreateDataChannel(expectedLabel, nil)
-					if createErr != nil {
-						t.Fatalf("Failed to create a PC pair for testing")
-					}
+					assert.NoError(t, createErr)
 
 					assert.True(t, dc.Ordered(), "Ordered should be set to true")
 
-					dc.OnMessage(func(msg DataChannelMessage) {
+					dc.OnMessage(func(DataChannelMessage) {
 						done <- true
 					})
 
 					if e := dc.SendText("Ping"); e != nil {
 						// wasm binding doesn't fire OnOpen (we probably already missed it)
 						dc.OnOpen(func() {
-							e = dc.SendText("Ping")
-							if e != nil {
-								t.Fatalf("Failed to send string on data channel")
-							}
+							assert.NoError(t, dc.SendText("Ping"), "Failed to send string on data channel")
 						})
 					}
 				})
@@ -324,9 +303,7 @@ func TestDataChannel_Send(t *testing.T) {
 		})
 
 		err = signalPair(offerPC, answerPC)
-		if err != nil {
-			t.Fatalf("Failed to signal our PC pair for testing")
-		}
+		assert.NoError(t, err)
 
 		closePair(t, offerPC, answerPC, done)
 	})
@@ -359,7 +336,7 @@ func TestDataChannel_Close(t *testing.T) {
 	})
 }
 
-func TestDataChannelParameters(t *testing.T) {
+func TestDataChannelParameters(t *testing.T) { //nolint:cyclop
 	report := test.CheckRoutines(t)
 	defer report()
 
@@ -468,37 +445,29 @@ func TestDataChannelParameters(t *testing.T) {
 
 		answerPC.OnDataChannel(func(d *DataChannel) {
 			// Ignore our default channel, exists to force ICE candidates. See signalPair for more info
-			if d.Label() == "initial_data_channel" {
-				return
-			}
-
-			t.Fatal("OnDataChannel must not be fired when negotiated == true")
+			assert.Equal(t, "initial_data_channel", d.Label(), "OnDataChannel must not be fired when negotiated == true")
 		})
-		offerPC.OnDataChannel(func(d *DataChannel) {
-			t.Fatal("OnDataChannel must not be fired when negotiated == true")
+		offerPC.OnDataChannel(func(*DataChannel) {
+			assert.Fail(t, "OnDataChannel must not be fired when negotiated == true")
 		})
 
-		seenAnswerMessage := &atomicBool{}
-		seenOfferMessage := &atomicBool{}
+		seenAnswerMessage := &atomic.Bool{}
+		seenOfferMessage := &atomic.Bool{}
 
 		answerDatachannel.OnMessage(func(msg DataChannelMessage) {
 			if msg.IsString && string(msg.Data) == expectedMessage {
-				seenAnswerMessage.set(true)
+				seenAnswerMessage.Store(true)
 			}
 		})
 
 		offerDatachannel.OnMessage(func(msg DataChannelMessage) {
 			if msg.IsString && string(msg.Data) == expectedMessage {
-				seenOfferMessage.set(true)
+				seenOfferMessage.Store(true)
 			}
 		})
 
 		go func() {
-			for {
-				if seenAnswerMessage.get() && seenOfferMessage.get() {
-					break
-				}
-
+			for seenAnswerMessage.Load() && seenOfferMessage.Load() {
 				if offerDatachannel.ReadyState() == DataChannelStateOpen {
 					assert.NoError(t, offerDatachannel.SendText(expectedMessage))
 				}
