@@ -1,6 +1,10 @@
+// SPDX-FileCopyrightText: 2023 The Pion community <https://pion.ly>
+// SPDX-License-Identifier: MIT
+
 //go:build !js
 // +build !js
 
+// ice-single-port demonstrates Pion WebRTC's ability to serve many PeerConnections on a single port.
 package main
 
 import (
@@ -9,14 +13,14 @@ import (
 	"net/http"
 	"time"
 
-	"github.com/pion/ice/v2"
-	"github.com/pion/webrtc/v3"
+	"github.com/pion/ice/v4"
+	"github.com/pion/webrtc/v4"
 )
 
 var api *webrtc.API //nolint
 
 // Everything below is the Pion WebRTC API! Thanks for using it ❤️.
-func doSignaling(w http.ResponseWriter, r *http.Request) {
+func doSignaling(res http.ResponseWriter, req *http.Request) {
 	peerConnection, err := api.NewPeerConnection(webrtc.Configuration{})
 	if err != nil {
 		panic(err)
@@ -40,7 +44,7 @@ func doSignaling(w http.ResponseWriter, r *http.Request) {
 	})
 
 	var offer webrtc.SessionDescription
-	if err = json.NewDecoder(r.Body).Decode(&offer); err != nil {
+	if err = json.NewDecoder(req.Body).Decode(&offer); err != nil {
 		panic(err)
 	}
 
@@ -68,8 +72,8 @@ func doSignaling(w http.ResponseWriter, r *http.Request) {
 		panic(err)
 	}
 
-	w.Header().Set("Content-Type", "application/json")
-	if _, err := w.Write(response); err != nil {
+	res.Header().Set("Content-Type", "application/json")
+	if _, err := res.Write(response); err != nil {
 		panic(err)
 	}
 }
@@ -96,5 +100,6 @@ func main() {
 	http.HandleFunc("/doSignaling", doSignaling)
 
 	fmt.Println("Open http://localhost:8080 to access this demo")
+	// nolint: gosec
 	panic(http.ListenAndServe(":8080", nil))
 }
