@@ -1,3 +1,6 @@
+// SPDX-FileCopyrightText: 2023 The Pion community <https://pion.ly>
+// SPDX-License-Identifier: MIT
+
 //go:build !js
 // +build !js
 
@@ -9,7 +12,7 @@ import (
 	"testing"
 	"time"
 
-	"github.com/pion/transport/v2/test"
+	"github.com/pion/transport/v3/test"
 	"github.com/stretchr/testify/assert"
 )
 
@@ -72,16 +75,17 @@ func TestICETransport_OnSelectedCandidatePairChange(t *testing.T) {
 	})
 
 	senderCalledCandidateChange := int32(0)
-	pcOffer.SCTP().Transport().ICETransport().OnSelectedCandidatePairChange(func(pair *ICECandidatePair) {
+	pcOffer.SCTP().Transport().ICETransport().OnSelectedCandidatePairChange(func(*ICECandidatePair) {
 		atomic.StoreInt32(&senderCalledCandidateChange, 1)
 	})
 
 	assert.NoError(t, signalPair(pcOffer, pcAnswer))
 	<-iceComplete
 
-	if atomic.LoadInt32(&senderCalledCandidateChange) == 0 {
-		t.Fatalf("Sender ICETransport OnSelectedCandidateChange was never called")
-	}
+	assert.NotEmpty(
+		t, atomic.LoadInt32(&senderCalledCandidateChange),
+		"Sender ICETransport OnSelectedCandidateChange was never called",
+	)
 
 	closePairNow(t, pcOffer, pcAnswer)
 }
@@ -95,10 +99,14 @@ func TestICETransport_GetSelectedCandidatePair(t *testing.T) {
 	offererSelectedPair, err := offerer.SCTP().Transport().ICETransport().GetSelectedCandidatePair()
 	assert.NoError(t, err)
 	assert.Nil(t, offererSelectedPair)
+	_, statsAvailable := offerer.SCTP().Transport().ICETransport().GetSelectedCandidatePairStats()
+	assert.False(t, statsAvailable)
 
 	answererSelectedPair, err := answerer.SCTP().Transport().ICETransport().GetSelectedCandidatePair()
 	assert.NoError(t, err)
 	assert.Nil(t, answererSelectedPair)
+	_, statsAvailable = answerer.SCTP().Transport().ICETransport().GetSelectedCandidatePairStats()
+	assert.False(t, statsAvailable)
 
 	assert.NoError(t, signalPair(offerer, answerer))
 	peerConnectionConnected.Wait()
@@ -106,10 +114,46 @@ func TestICETransport_GetSelectedCandidatePair(t *testing.T) {
 	offererSelectedPair, err = offerer.SCTP().Transport().ICETransport().GetSelectedCandidatePair()
 	assert.NoError(t, err)
 	assert.NotNil(t, offererSelectedPair)
+	_, statsAvailable = offerer.SCTP().Transport().ICETransport().GetSelectedCandidatePairStats()
+	assert.True(t, statsAvailable)
 
 	answererSelectedPair, err = answerer.SCTP().Transport().ICETransport().GetSelectedCandidatePair()
 	assert.NoError(t, err)
 	assert.NotNil(t, answererSelectedPair)
+	_, statsAvailable = answerer.SCTP().Transport().ICETransport().GetSelectedCandidatePairStats()
+	assert.True(t, statsAvailable)
+
+	closePairNow(t, offerer, answerer)
+}
+
+func TestICETransport_GetLocalAndRemoteParameters(t *testing.T) {
+	offerer, answerer, err := newPair()
+	assert.NoError(t, err)
+
+	_, err = offerer.SCTP().Transport().ICETransport().GetRemoteParameters()
+	assert.Error(t, err, errICEAgentNotExist)
+
+	peerConnectionConnected := untilConnectionState(PeerConnectionStateConnected, offerer, answerer)
+
+	assert.NoError(t, signalPair(offerer, answerer))
+	peerConnectionConnected.Wait()
+
+	offerLocalParameters, err := offerer.SCTP().Transport().ICETransport().GetLocalParameters()
+	assert.NoError(t, err)
+
+	offerRemoteParameters, err := offerer.SCTP().Transport().ICETransport().GetRemoteParameters()
+	assert.NoError(t, err)
+
+	answerLocalParameters, err := answerer.SCTP().Transport().ICETransport().GetLocalParameters()
+	assert.NoError(t, err)
+
+	answerRemoteParameters, err := answerer.SCTP().Transport().ICETransport().GetRemoteParameters()
+	assert.NoError(t, err)
+
+	assert.Equal(t, offerLocalParameters.UsernameFragment, answerRemoteParameters.UsernameFragment)
+	assert.Equal(t, offerLocalParameters.Password, answerRemoteParameters.Password)
+	assert.Equal(t, answerLocalParameters.UsernameFragment, offerRemoteParameters.UsernameFragment)
+	assert.Equal(t, answerLocalParameters.Password, offerRemoteParameters.Password)
 
 	closePairNow(t, offerer, answerer)
 }
