@@ -1,3 +1,6 @@
+// SPDX-FileCopyrightText: 2023 The Pion community <https://pion.ly>
+// SPDX-License-Identifier: MIT
+
 package webrtc
 
 import (
@@ -11,7 +14,7 @@ func TestNetworkType_String(t *testing.T) {
 		cType          NetworkType
 		expectedString string
 	}{
-		{NetworkType(Unknown), unknownStr},
+		{NetworkTypeUnknown, ErrUnknownType.Error()},
 		{NetworkTypeUDP4, "udp4"},
 		{NetworkTypeUDP6, "udp6"},
 		{NetworkTypeTCP4, "tcp4"},
@@ -33,7 +36,7 @@ func TestNetworkType(t *testing.T) {
 		shouldFail   bool
 		expectedType NetworkType
 	}{
-		{unknownStr, true, NetworkType(Unknown)},
+		{ErrUnknownType.Error(), true, NetworkTypeUnknown},
 		{"udp4", false, NetworkTypeUDP4},
 		{"udp6", false, NetworkTypeUDP6},
 		{"tcp4", false, NetworkTypeTCP4},
@@ -42,8 +45,10 @@ func TestNetworkType(t *testing.T) {
 
 	for i, testCase := range testCases {
 		actual, err := NewNetworkType(testCase.typeString)
-		if (err != nil) != testCase.shouldFail {
-			t.Error(err)
+		if testCase.shouldFail {
+			assert.Error(t, err)
+		} else {
+			assert.NoError(t, err)
 		}
 		assert.Equal(t,
 			testCase.expectedType,
