@@ -1,3 +1,6 @@
+// SPDX-FileCopyrightText: 2023 The Pion community <https://pion.ly>
+// SPDX-License-Identifier: MIT
+
 //go:build !js
 // +build !js
 
@@ -8,8 +11,8 @@ import (
 	"testing"
 	"time"
 
-	"github.com/pion/transport/v2/test"
-	"github.com/pion/webrtc/v3/pkg/media"
+	"github.com/pion/transport/v3/test"
+	"github.com/pion/webrtc/v4/pkg/media"
 	"github.com/stretchr/testify/assert"
 )
 
@@ -23,9 +26,6 @@ func Test_ORTC_Media(t *testing.T) {
 	stackA, stackB, err := newORTCPair()
 	assert.NoError(t, err)
 
-	assert.NoError(t, stackA.api.mediaEngine.RegisterDefaultCodecs())
-	assert.NoError(t, stackB.api.mediaEngine.RegisterDefaultCodecs())
-
 	assert.NoError(t, signalORTCPair(stackA, stackB))
 
 	track, err := NewTrackLocalStaticSample(RTPCodecCapability{MimeType: MimeTypeVP8}, "video", "pion")
