@@ -1,3 +1,6 @@
+// SPDX-FileCopyrightText: 2023 The Pion community <https://pion.ly>
+// SPDX-License-Identifier: MIT
+
 package webrtc
 
 import (
@@ -11,7 +14,7 @@ func TestNewPeerConnectionState(t *testing.T) {
 		stateString   string
 		expectedState PeerConnectionState
 	}{
-		{unknownStr, PeerConnectionState(Unknown)},
+		{ErrUnknownType.Error(), PeerConnectionStateUnknown},
 		{"new", PeerConnectionStateNew},
 		{"connecting", PeerConnectionStateConnecting},
 		{"connected", PeerConnectionStateConnected},
@@ -34,7 +37,7 @@ func TestPeerConnectionState_String(t *testing.T) {
 		state          PeerConnectionState
 		expectedString string
 	}{
-		{PeerConnectionState(Unknown), unknownStr},
+		{PeerConnectionStateUnknown, ErrUnknownType.Error()},
 		{PeerConnectionStateNew, "new"},
 		{PeerConnectionStateConnecting, "connecting"},
 		{PeerConnectionStateConnected, "connected"},
