@@ -1,12 +1,18 @@
+// SPDX-FileCopyrightText: 2023 The Pion community <https://pion.ly>
+// SPDX-License-Identifier: MIT
+
 package samplebuilder
 
 import (
 	"fmt"
+	"runtime"
+	"slices"
+	"sync/atomic"
 	"testing"
 	"time"
 
 	"github.com/pion/rtp"
-	"github.com/pion/webrtc/v3/pkg/media"
+	"github.com/pion/webrtc/v4/pkg/media"
 	"github.com/stretchr/testify/assert"
 )
 
@@ -14,6 +20,7 @@ type sampleBuilderTest struct {
 	message          string
 	packets          []*rtp.Packet
 	withHeadChecker  bool
+	withRTPHeader    bool
 	headBytes        []byte
 	samples          []*media.Sample
 	maxLate          uint16
@@ -23,6 +30,7 @@ type sampleBuilderTest struct {
 type fakeDepacketizer struct {
 	headChecker bool
 	headBytes   []byte
+	alwaysHead  bool
 }
 
 func (f *fakeDepacketizer) Unmarshal(r []byte) ([]byte, error) {
@@ -35,19 +43,24 @@ func (f *fakeDepacketizer) IsPartitionHead(payload []byte) bool {
 		// the tests should be fixed to not assume the bug
 		return true
 	}
-	for _, b := range f.headBytes {
-		if payload[0] == b {
-			return true
-		}
+
+	// skip padding
+	if len(payload) < 1 {
+		return false
 	}
-	return false
+
+	if f.alwaysHead {
+		return true
+	}
+
+	return slices.Contains(f.headBytes, payload[0])
 }
 
-func (f *fakeDepacketizer) IsPartitionTail(marker bool, payload []byte) bool {
+func (f *fakeDepacketizer) IsPartitionTail(marker bool, _ []byte) bool {
 	return marker
 }
 
-func TestSampleBuilder(t *testing.T) {
+func TestSampleBuilder(t *testing.T) { //nolint:maintidx
 	testData := []sampleBuilderTest{
 		{
 			message: "SampleBuilder shouldn't emit anything if only one RTP packet has been pushed",
@@ -59,6 +72,7 @@ func TestSampleBuilder(t *testing.T) {
 			maxLateTimestamp: 0,
 		},
 		{
+			//nolint:lll
 			message: "SampleBuilder shouldn't emit anything if only one RTP packet has been pushed even if the market bit is set",
 			packets: []*rtp.Packet{
 				{Header: rtp.Header{SequenceNumber: 5000, Timestamp: 5, Marker: true}, Payload: []byte{0x01}},
@@ -223,57 +237,151 @@ func TestSampleBuilder(t *testing.T) {
 			maxLate:          50,
 			maxLateTimestamp: 2000,
 		},
+		{
+			message: "Sample builder should recognize padding packets",
+			packets: []*rtp.Packet{
+				// 1st packet
+				{Header: rtp.Header{SequenceNumber: 5000, Timestamp: 1}, Payload: []byte{1}},
+				// 2nd packet
+				{Header: rtp.Header{SequenceNumber: 5001, Timestamp: 1}, Payload: []byte{2}},
+				// 3rd packet
+				{Header: rtp.Header{SequenceNumber: 5002, Timestamp: 1, Marker: true}, Payload: []byte{3}},
+				// Padding packet 1
+				{Header: rtp.Header{SequenceNumber: 5003, Timestamp: 1}, Payload: []byte{}},
+				// Padding packet 2
+				{Header: rtp.Header{SequenceNumber: 5004, Timestamp: 1}, Payload: []byte{}},
+				// 6th packet
+				{Header: rtp.Header{SequenceNumber: 5005, Timestamp: 3}, Payload: []byte{1}},
+				// 7th packet
+				{Header: rtp.Header{SequenceNumber: 5006, Timestamp: 3, Marker: true}, Payload: []byte{7}},
+				// 7th packet
+				{Header: rtp.Header{SequenceNumber: 5007, Timestamp: 4}, Payload: []byte{1}},
+			},
+			withHeadChecker: true,
+			headBytes:       []byte{1},
+			samples: []*media.Sample{
+				{Data: []byte{1, 2, 3}, Duration: 0, PacketTimestamp: 1, PrevDroppedPackets: 0}, // first sample
+			},
+			maxLate:          50,
+			maxLateTimestamp: 2000,
+		},
+		{
+			//nolint:lll
+			message: "Sample builder should build a sample out of a packet that's both start and end following a run of padding packets",
+			packets: []*rtp.Packet{
+				// 1st valid packet
+				{Header: rtp.Header{SequenceNumber: 5000, Timestamp: 1}, Payload: []byte{1}},
+				// 2nd valid packet
+				{Header: rtp.Header{SequenceNumber: 5001, Timestamp: 1, Marker: true}, Payload: []byte{2}},
+				// 1st padding packet
+				{Header: rtp.Header{SequenceNumber: 5002, Timestamp: 1}, Payload: []byte{}},
+				// 2nd padding packet
+				{Header: rtp.Header{SequenceNumber: 5003, Timestamp: 1}, Payload: []byte{}},
+				// 3rd valid packet
+				{Header: rtp.Header{SequenceNumber: 5004, Timestamp: 2, Marker: true}, Payload: []byte{1}},
+				// 4th valid packet, start of next sample
+				{Header: rtp.Header{SequenceNumber: 5005, Timestamp: 3}, Payload: []byte{1}},
+			},
+			withHeadChecker: true,
+			headBytes:       []byte{1},
+			samples: []*media.Sample{
+				{Data: []byte{1, 2}, Duration: 0, PacketTimestamp: 1, PrevDroppedPackets: 0}, // 1st sample
+			},
+			maxLate:          50,
+			maxLateTimestamp: 2000,
+		},
+		{
+			message: "SampleBuilder should emit samples with RTP headers when WithRTPHeaders option is enabled",
+			packets: []*rtp.Packet{
+				{Header: rtp.Header{SequenceNumber: 5000, Timestamp: 5}, Payload: []byte{0x01}},
+				{Header: rtp.Header{SequenceNumber: 5001, Timestamp: 6}, Payload: []byte{0x02}},
+				{Header: rtp.Header{SequenceNumber: 5002, Timestamp: 6}, Payload: []byte{0x03}},
+				{Header: rtp.Header{SequenceNumber: 5003, Timestamp: 7}, Payload: []byte{0x04}},
+			},
+			samples: []*media.Sample{
+				{Data: []byte{0x01}, Duration: time.Second, PacketTimestamp: 5, RTPHeaders: []*rtp.Header{
+					{SequenceNumber: 5000, Timestamp: 5},
+				}},
+				{Data: []byte{0x02, 0x03}, Duration: time.Second, PacketTimestamp: 6, RTPHeaders: []*rtp.Header{
+					{SequenceNumber: 5001, Timestamp: 6},
+					{SequenceNumber: 5002, Timestamp: 6},
+				}},
+			},
+			maxLate:          50,
+			maxLateTimestamp: 0,
+			withRTPHeader:    true,
+		},
 	}
 
 	t.Run("Pop", func(t *testing.T) {
 		assert := assert.New(t)
 
-		for _, t := range testData {
+		for _, td := range testData {
 			var opts []Option
-			if t.maxLateTimestamp != 0 {
+			if td.maxLateTimestamp != 0 {
 				opts = append(opts, WithMaxTimeDelay(
-					time.Millisecond*time.Duration(int64(t.maxLateTimestamp)),
+					time.Millisecond*time.Duration(int64(td.maxLateTimestamp)),
 				))
 			}
+			if td.withRTPHeader {
+				opts = append(opts, WithRTPHeaders(true))
+			}
 
 			d := &fakeDepacketizer{
-				headChecker: t.withHeadChecker,
-				headBytes:   t.headBytes,
+				headChecker: td.withHeadChecker,
+				headBytes:   td.headBytes,
 			}
-			s := New(t.maxLate, d, 1, opts...)
+			s := New(td.maxLate, d, 1, opts...)
 			samples := []*media.Sample{}
 
-			for _, p := range t.packets {
+			for _, p := range td.packets {
 				s.Push(p)
 			}
 			for sample := s.Pop(); sample != nil; sample = s.Pop() {
 				samples = append(samples, sample)
 			}
-			assert.Equal(t.samples, samples, t.message)
+			assert.Equal(td.samples, samples, td.message)
 		}
 	})
 }
 
-// SampleBuilder should respect maxLate if we popped successfully but then have a gap larger then maxLate
+// SampleBuilder should respect maxLate if we popped successfully but then have a gap larger then maxLate.
 func TestSampleBuilderMaxLate(t *testing.T) {
 	assert := assert.New(t)
-	s := New(50, &fakeDepacketizer{}, 1)
+	fd := New(50, &fakeDepacketizer{}, 1)
 
-	s.Push(&rtp.Packet{Header: rtp.Header{SequenceNumber: 0, Timestamp: 1}, Payload: []byte{0x01}})
-	s.Push(&rtp.Packet{Header: rtp.Header{SequenceNumber: 1, Timestamp: 2}, Payload: []byte{0x01}})
-	s.Push(&rtp.Packet{Header: rtp.Header{SequenceNumber: 2, Timestamp: 3}, Payload: []byte{0x01}})
-	assert.Equal(&media.Sample{Data: []byte{0x01}, Duration: time.Second, PacketTimestamp: 1}, s.Pop(), "Failed to build samples before gap")
+	fd.Push(&rtp.Packet{Header: rtp.Header{SequenceNumber: 0, Timestamp: 1}, Payload: []byte{0x01}})
+	fd.Push(&rtp.Packet{Header: rtp.Header{SequenceNumber: 1, Timestamp: 2}, Payload: []byte{0x01}})
+	fd.Push(&rtp.Packet{Header: rtp.Header{SequenceNumber: 2, Timestamp: 3}, Payload: []byte{0x01}})
+	assert.Equal(&media.Sample{
+		Data:            []byte{0x01},
+		Duration:        time.Second,
+		PacketTimestamp: 1,
+	}, fd.Pop(), "Failed to build samples before gap")
 
-	s.Push(&rtp.Packet{Header: rtp.Header{SequenceNumber: 5000, Timestamp: 500}, Payload: []byte{0x02}})
-	s.Push(&rtp.Packet{Header: rtp.Header{SequenceNumber: 5001, Timestamp: 501}, Payload: []byte{0x02}})
-	s.Push(&rtp.Packet{Header: rtp.Header{SequenceNumber: 5002, Timestamp: 502}, Payload: []byte{0x02}})
+	fd.Push(&rtp.Packet{Header: rtp.Header{SequenceNumber: 5000, Timestamp: 500}, Payload: []byte{0x02}})
+	fd.Push(&rtp.Packet{Header: rtp.Header{SequenceNumber: 5001, Timestamp: 501}, Payload: []byte{0x02}})
+	fd.Push(&rtp.Packet{Header: rtp.Header{SequenceNumber: 5002, Timestamp: 502}, Payload: []byte{0x02}})
 
-	assert.Equal(&media.Sample{Data: []byte{0x01}, Duration: time.Second, PacketTimestamp: 2}, s.Pop(), "Failed to build samples after large gap")
-	assert.Equal((*media.Sample)(nil), s.Pop(), "Failed to build samples after large gap")
+	assert.Equal(&media.Sample{
+		Data:            []byte{0x01},
+		Duration:        time.Second,
+		PacketTimestamp: 2,
+	}, fd.Pop(), "Failed to build samples after large gap")
+	assert.Equal((*media.Sample)(nil), fd.Pop(), "Failed to build samples after large gap")
 
-	s.Push(&rtp.Packet{Header: rtp.Header{SequenceNumber: 6000, Timestamp: 600}, Payload: []byte{0x03}})
-	assert.Equal(&media.Sample{Data: []byte{0x02}, Duration: time.Second, PacketTimestamp: 500, PrevDroppedPackets: 4998}, s.Pop(), "Failed to build samples after large gap")
-	assert.Equal(&media.Sample{Data: []byte{0x02}, Duration: time.Second, PacketTimestamp: 501}, s.Pop(), "Failed to build samples after large gap")
+	fd.Push(&rtp.Packet{Header: rtp.Header{SequenceNumber: 6000, Timestamp: 600}, Payload: []byte{0x03}})
+	assert.Equal(&media.Sample{
+		Data:               []byte{0x02},
+		Duration:           time.Second,
+		PacketTimestamp:    500,
+		PrevDroppedPackets: 4998,
+	}, fd.Pop(), "Failed to build samples after large gap")
+	assert.Equal(&media.Sample{
+		Data:            []byte{0x02},
+		Duration:        time.Second,
+		PacketTimestamp: 501,
+	}, fd.Pop(), "Failed to build samples after large gap")
 }
 
 func TestSeqnumDistance(t *testing.T) {
@@ -291,10 +399,7 @@ func TestSeqnumDistance(t *testing.T) {
 	}
 
 	for _, data := range testData {
-		if ret := seqnumDistance(data.x, data.y); ret != data.d {
-			t.Errorf("seqnumDistance(%d, %d) returned %d which must be %d",
-				data.x, data.y, ret, data.d)
-		}
+		assert.Equalf(t, data.d, seqnumDistance(data.x, data.y), "seqnumDistance(%d, %d)", data.x, data.y)
 	}
 }
 
@@ -306,27 +411,30 @@ func TestSampleBuilderCleanReference(t *testing.T) {
 	} {
 		seqStart := seqStart
 		t.Run(fmt.Sprintf("From%d", seqStart), func(t *testing.T) {
-			s := New(10, &fakeDepacketizer{}, 1)
+			fd := New(10, &fakeDepacketizer{}, 1)
 
-			s.Push(&rtp.Packet{Header: rtp.Header{SequenceNumber: 0 + seqStart, Timestamp: 0}, Payload: []byte{0x01}})
-			s.Push(&rtp.Packet{Header: rtp.Header{SequenceNumber: 1 + seqStart, Timestamp: 0}, Payload: []byte{0x02}})
-			s.Push(&rtp.Packet{Header: rtp.Header{SequenceNumber: 2 + seqStart, Timestamp: 0}, Payload: []byte{0x03}})
+			fd.Push(&rtp.Packet{Header: rtp.Header{SequenceNumber: 0 + seqStart, Timestamp: 0}, Payload: []byte{0x01}})
+			fd.Push(&rtp.Packet{Header: rtp.Header{SequenceNumber: 1 + seqStart, Timestamp: 0}, Payload: []byte{0x02}})
+			fd.Push(&rtp.Packet{Header: rtp.Header{SequenceNumber: 2 + seqStart, Timestamp: 0}, Payload: []byte{0x03}})
 			pkt4 := &rtp.Packet{Header: rtp.Header{SequenceNumber: 14 + seqStart, Timestamp: 120}, Payload: []byte{0x04}}
-			s.Push(pkt4)
+			fd.Push(pkt4)
 			pkt5 := &rtp.Packet{Header: rtp.Header{SequenceNumber: 12 + seqStart, Timestamp: 120}, Payload: []byte{0x05}}
-			s.Push(pkt5)
+			fd.Push(pkt5)
 
 			for i := 0; i < 3; i++ {
-				if s.buffer[(i+int(seqStart))%0x10000] != nil {
-					t.Errorf("Old packet (%d) is not unreferenced (maxLate: 10, pushed: 12)", i)
-				}
-			}
-			if s.buffer[(14+int(seqStart))%0x10000] != pkt4 {
-				t.Error("New packet must be referenced after jump")
-			}
-			if s.buffer[(12+int(seqStart))%0x10000] != pkt5 {
-				t.Error("New packet must be referenced after jump")
+				assert.Nilf(
+					t, fd.buffer[(i+int(seqStart))%0x10000],
+					"Old packet (%d) is not unreferenced (maxLate: 10, pushed: 12)", i,
+				)
 			}
+			assert.Equal(
+				t, pkt4, fd.buffer[(14+int(seqStart))%0x10000],
+				"New packet must be referenced after jump",
+			)
+			assert.Equal(
+				t, pkt5, fd.buffer[(12+int(seqStart))%0x10000],
+				"New packet must be referenced after jump",
+			)
 		})
 	}
 }
@@ -343,9 +451,7 @@ func TestSampleBuilderPushMaxZero(t *testing.T) {
 
 	s := New(0, d, 1)
 	s.Push(&pkts[0])
-	if sample := s.Pop(); sample == nil {
-		t.Error("Should expect a popped sample")
-	}
+	assert.NotNil(t, s.Pop(), "Should expect a sample")
 }
 
 func TestSampleBuilderWithPacketReleaseHandler(t *testing.T) {
@@ -362,206 +468,291 @@ func TestSampleBuilderWithPacketReleaseHandler(t *testing.T) {
 		{Header: rtp.Header{SequenceNumber: 13, Timestamp: 122}, Payload: []byte{0x04}},
 		{Header: rtp.Header{SequenceNumber: 21, Timestamp: 200}, Payload: []byte{0x05}},
 	}
-	s := New(10, &fakeDepacketizer{}, 1, WithPacketReleaseHandler(fakePacketReleaseHandler))
-	s.Push(&pkts[0])
-	s.Push(&pkts[1])
-	if len(released) == 0 {
-		t.Errorf("Old packet is not released")
-	}
-	if len(released) > 0 && released[0].SequenceNumber != pkts[0].SequenceNumber {
-		t.Errorf("Unexpected packet released by maxLate")
-	}
+	fd := New(10, &fakeDepacketizer{}, 1, WithPacketReleaseHandler(fakePacketReleaseHandler))
+	fd.Push(&pkts[0])
+	fd.Push(&pkts[1])
+	assert.NotEmpty(t, released, "Old packet is not released")
+	assert.Equal(t, pkts[0].SequenceNumber, released[0].SequenceNumber, "Unexpected packet released by maxLate")
 	// Test packets released after samples built.
-	s.Push(&pkts[2])
-	s.Push(&pkts[3])
-	s.Push(&pkts[4])
-	if s.Pop() == nil {
-		t.Errorf("Should have some sample here.")
-	}
-	if len(released) < 3 {
-		t.Errorf("packet built with sample is not released")
-	}
-	if len(released) >= 2 && released[2].SequenceNumber != pkts[2].SequenceNumber {
-		t.Errorf("Unexpected packet released by samples built")
-	}
+	fd.Push(&pkts[2])
+	fd.Push(&pkts[3])
+	fd.Push(&pkts[4])
+	assert.NotNil(t, fd.Pop(), "Should have some sample here.")
+	assert.GreaterOrEqual(t, len(released), 3, "packet built with sample is not released")
+	assert.Equal(t, pkts[2].SequenceNumber, released[2].SequenceNumber, "Unexpected packet released by samples built")
 }
 
-func TestPopWithTimestamp(t *testing.T) {
-	t.Run("Crash on nil", func(t *testing.T) {
-		s := New(0, &fakeDepacketizer{}, 1)
-		sample, timestamp := s.PopWithTimestamp()
-		assert.Nil(t, sample)
-		assert.Equal(t, uint32(0), timestamp)
-	})
-}
+func TestSampleBuilderWithPacketHeadHandler(t *testing.T) {
+	packets := []*rtp.Packet{
+		{Header: rtp.Header{SequenceNumber: 5000, Timestamp: 5}, Payload: []byte{0x01}},
+		{Header: rtp.Header{SequenceNumber: 5001, Timestamp: 5}, Payload: []byte{0x02}},
+		{Header: rtp.Header{SequenceNumber: 5002, Timestamp: 6}, Payload: []byte{0x01}},
+		{Header: rtp.Header{SequenceNumber: 5003, Timestamp: 6}, Payload: []byte{0x02}},
+		{Header: rtp.Header{SequenceNumber: 5004, Timestamp: 7}, Payload: []byte{0x01}},
+	}
 
-type truePartitionHeadChecker struct{}
+	headCount := 0
+	s := New(10, &fakeDepacketizer{}, 1, WithPacketHeadHandler(func(any) any {
+		headCount++
 
-func (f *truePartitionHeadChecker) IsPartitionHead(payload []byte) bool {
-	return true
+		return true
+	}))
+
+	for _, pkt := range packets {
+		s.Push(pkt)
+	}
+
+	for {
+		sample := s.Pop()
+		if sample == nil {
+			break
+		}
+
+		assert.NotNil(t, sample.Metadata, "sample metadata shouldn't be nil")
+		assert.Equal(t, true, sample.Metadata, "sample metadata should've been set to true")
+	}
+
+	assert.Equal(t, 2, headCount, "two sample heads should have been inspected")
 }
 
 func TestSampleBuilderData(t *testing.T) {
-	s := New(10, &fakeDepacketizer{}, 1,
-		WithPartitionHeadChecker(&truePartitionHeadChecker{}),
-	)
-	j := 0
+	fd := New(10, &fakeDepacketizer{
+		headChecker: true,
+		alwaysHead:  true,
+	}, 1)
+	validSamples := 0
 	for i := 0; i < 0x20000; i++ {
-		p := rtp.Packet{
+		packet := rtp.Packet{
 			Header: rtp.Header{
-				SequenceNumber: uint16(i),
-				Timestamp:      uint32(i + 42),
+				SequenceNumber: uint16(i),      //nolint:gosec // G115
+				Timestamp:      uint32(i + 42), //nolint:gosec // G115
 			},
 			Payload: []byte{byte(i)},
 		}
-		s.Push(&p)
+		fd.Push(&packet)
 		for {
-			sample, ts := s.PopWithTimestamp()
+			sample := fd.Pop()
 			if sample == nil {
 				break
 			}
-			assert.Equal(t, ts, uint32(j+42), "timestamp")
+			assert.Equal(t, sample.PacketTimestamp, uint32(validSamples+42), "timestamp") //nolint:gosec // G115
 			assert.Equal(t, len(sample.Data), 1, "data length")
-			assert.Equal(t, byte(j), sample.Data[0], "data")
-			j++
+			assert.Equal(t, byte(validSamples), sample.Data[0], "data")
+			validSamples++
 		}
 	}
 	// only the last packet should be dropped
-	assert.Equal(t, j, 0x1FFFF)
+	assert.Equal(t, validSamples, 0x1FFFF)
+}
+
+func TestSampleBuilderPacketUnreference(t *testing.T) {
+	fd := New(10, &fakeDepacketizer{
+		headChecker: true,
+	}, 1)
+
+	var refs int64
+	finalizer := func(*rtp.Packet) {
+		atomic.AddInt64(&refs, -1)
+	}
+
+	for i := 0; i < 0x20000; i++ {
+		atomic.AddInt64(&refs, 1)
+		packet := rtp.Packet{
+			Header: rtp.Header{
+				SequenceNumber: uint16(i),      //nolint:gosec // G115
+				Timestamp:      uint32(i + 42), //nolint:gosec // G115
+			},
+			Payload: []byte{byte(i)},
+		}
+		runtime.SetFinalizer(&packet, finalizer)
+		fd.Push(&packet)
+		for {
+			sample := fd.Pop()
+			if sample == nil {
+				break
+			}
+		}
+	}
+
+	runtime.GC()
+	time.Sleep(10 * time.Millisecond)
+
+	remainedRefs := atomic.LoadInt64(&refs)
+	runtime.KeepAlive(fd)
+
+	// only the last packet should be still referenced
+	assert.Equal(t, int64(1), remainedRefs)
+}
+
+func TestSampleBuilder_Flush(t *testing.T) {
+	fd := New(50, &fakeDepacketizer{
+		headChecker: true,
+		headBytes:   []byte{0x01},
+	}, 1)
+
+	fd.Push(&rtp.Packet{
+		Header:  rtp.Header{SequenceNumber: 999, Timestamp: 0},
+		Payload: []byte{0x00},
+	}) // Invalid packet
+	// Gap preventing below packets to be processed
+	fd.Push(&rtp.Packet{
+		Header:  rtp.Header{SequenceNumber: 1001, Timestamp: 1, Marker: true},
+		Payload: []byte{0x01, 0x11},
+	}) // Valid packet
+	fd.Push(&rtp.Packet{
+		Header:  rtp.Header{SequenceNumber: 1011, Timestamp: 10, Marker: true},
+		Payload: []byte{0x01, 0x12},
+	}) // Valid packet
+
+	assert.Nil(t, fd.Pop(), "Unexpected sample is returned. Test precondition may be broken")
+
+	fd.Flush()
+
+	samples := []*media.Sample{}
+	for sample := fd.Pop(); sample != nil; sample = fd.Pop() {
+		samples = append(samples, sample)
+	}
+
+	expected := []*media.Sample{
+		{Data: []byte{0x01, 0x11}, Duration: 9 * time.Second, PacketTimestamp: 1, PrevDroppedPackets: 2},
+		{Data: []byte{0x01, 0x12}, Duration: 0, PacketTimestamp: 10, PrevDroppedPackets: 9},
+	}
+
+	assert.Equal(t, expected, samples)
 }
 
 func BenchmarkSampleBuilderSequential(b *testing.B) {
-	s := New(100, &fakeDepacketizer{}, 1)
+	fd := New(100, &fakeDepacketizer{}, 1)
 	b.ResetTimer()
-	j := 0
+	validSamples := 0
 	for i := 0; i < b.N; i++ {
-		p := rtp.Packet{
+		packet := rtp.Packet{
 			Header: rtp.Header{
-				SequenceNumber: uint16(i),
-				Timestamp:      uint32(i + 42),
+				SequenceNumber: uint16(i),      //nolint:gosec // G115
+				Timestamp:      uint32(i + 42), //nolint:gosec // G115
 			},
 			Payload: make([]byte, 50),
 		}
-		s.Push(&p)
+		fd.Push(&packet)
 		for {
-			s := s.Pop()
+			s := fd.Pop()
 			if s == nil {
 				break
 			}
-			j++
+			validSamples++
 		}
 	}
-	if b.N > 200 && j < b.N-100 {
-		b.Errorf("Got %v (N=%v)", j, b.N)
+	if b.N > 200 && validSamples < b.N-100 {
+		b.Errorf("Got %v (N=%v)", validSamples, b.N)
 	}
 }
 
 func BenchmarkSampleBuilderLoss(b *testing.B) {
-	s := New(100, &fakeDepacketizer{}, 1)
+	fd := New(100, &fakeDepacketizer{}, 1)
 	b.ResetTimer()
-	j := 0
+	validSamples := 0
 	for i := 0; i < b.N; i++ {
 		if i%13 == 0 {
 			continue
 		}
-		p := rtp.Packet{
+		packet := rtp.Packet{
 			Header: rtp.Header{
-				SequenceNumber: uint16(i),
-				Timestamp:      uint32(i + 42),
+				SequenceNumber: uint16(i),      //nolint:gosec // G115
+				Timestamp:      uint32(i + 42), //nolint:gosec // G115
 			},
 			Payload: make([]byte, 50),
 		}
-		s.Push(&p)
+		fd.Push(&packet)
 		for {
-			s := s.Pop()
+			s := fd.Pop()
 			if s == nil {
 				break
 			}
-			j++
+			validSamples++
 		}
 	}
-	if b.N > 200 && j < b.N/2-100 {
-		b.Errorf("Got %v (N=%v)", j, b.N)
+	if b.N > 200 && validSamples < b.N/2-100 {
+		b.Errorf("Got %v (N=%v)", validSamples, b.N)
 	}
 }
 
 func BenchmarkSampleBuilderReordered(b *testing.B) {
-	s := New(100, &fakeDepacketizer{}, 1)
+	fd := New(100, &fakeDepacketizer{}, 1)
 	b.ResetTimer()
-	j := 0
+	validSamples := 0
 	for i := 0; i < b.N; i++ {
-		p := rtp.Packet{
+		packet := rtp.Packet{
 			Header: rtp.Header{
-				SequenceNumber: uint16(i ^ 3),
-				Timestamp:      uint32((i ^ 3) + 42),
+				SequenceNumber: uint16(i ^ 3),        //nolint:gosec // G115
+				Timestamp:      uint32((i ^ 3) + 42), //nolint:gosec // G115
 			},
 			Payload: make([]byte, 50),
 		}
-		s.Push(&p)
+		fd.Push(&packet)
 		for {
-			s := s.Pop()
+			s := fd.Pop()
 			if s == nil {
 				break
 			}
-			j++
+			validSamples++
 		}
 	}
-	if b.N > 2 && j < b.N-5 && j > b.N {
-		b.Errorf("Got %v (N=%v)", j, b.N)
+	if b.N > 2 && validSamples < b.N-5 && validSamples > b.N {
+		b.Errorf("Got %v (N=%v)", validSamples, b.N)
 	}
 }
 
 func BenchmarkSampleBuilderFragmented(b *testing.B) {
-	s := New(100, &fakeDepacketizer{}, 1)
+	fd := New(100, &fakeDepacketizer{}, 1)
 	b.ResetTimer()
-	j := 0
+	validSamples := 0
 	for i := 0; i < b.N; i++ {
-		p := rtp.Packet{
+		packet := rtp.Packet{
 			Header: rtp.Header{
-				SequenceNumber: uint16(i),
-				Timestamp:      uint32(i/2 + 42),
+				SequenceNumber: uint16(i),        //nolint:gosec // G115
+				Timestamp:      uint32(i/2 + 42), //nolint:gosec // G115
 			},
 			Payload: make([]byte, 50),
 		}
-		s.Push(&p)
+		fd.Push(&packet)
 		for {
-			s := s.Pop()
+			s := fd.Pop()
 			if s == nil {
 				break
 			}
-			j++
+			validSamples++
 		}
 	}
-	if b.N > 200 && j < b.N/2-100 {
-		b.Errorf("Got %v (N=%v)", j, b.N)
+	if b.N > 200 && validSamples < b.N/2-100 {
+		b.Errorf("Got %v (N=%v)", validSamples, b.N)
 	}
 }
 
 func BenchmarkSampleBuilderFragmentedLoss(b *testing.B) {
-	s := New(100, &fakeDepacketizer{}, 1)
+	fd := New(100, &fakeDepacketizer{}, 1)
 	b.ResetTimer()
-	j := 0
+	validSamples := 0
 	for i := 0; i < b.N; i++ {
 		if i%13 == 0 {
 			continue
 		}
-		p := rtp.Packet{
+		packet := rtp.Packet{
 			Header: rtp.Header{
-				SequenceNumber: uint16(i),
-				Timestamp:      uint32(i/2 + 42),
+				SequenceNumber: uint16(i),        //nolint:gosec // G115
+				Timestamp:      uint32(i/2 + 42), //nolint:gosec // G115
 			},
 			Payload: make([]byte, 50),
 		}
-		s.Push(&p)
+		fd.Push(&packet)
 		for {
-			s := s.Pop()
+			s := fd.Pop()
 			if s == nil {
 				break
 			}
-			j++
+			validSamples++
 		}
 	}
-	if b.N > 200 && j < b.N/3-100 {
-		b.Errorf("Got %v (N=%v)", j, b.N)
+	if b.N > 200 && validSamples < b.N/3-100 {
+		b.Errorf("Got %v (N=%v)", validSamples, b.N)
 	}
 }
